import { AttributePath, FunctionExpression, MathematicalExpression, UpdateExpression, ExpressionAttributes, serializeConditionExpression, serializeProjectionExpression } from '@aws/dynamodb-expressions';
import { BinarySet, Marshaller, NumberValueSet } from '@aws/dynamodb-auto-marshaller';

/**
 * An error thrown when a marshaller or unmarshaller cannot understand a node of
 * the provided schema.
 */
class InvalidSchemaError extends Error {
    constructor(node, message) {
        super(message);
        this.node = node;
    }
}

/**
 * An error thrown by the marshaller when a node of the provided input cannot be
 * marshalled into the type specified in the schema.
 */
class InvalidValueError extends Error {
    constructor(invalidValue, message) {
        super(message);
        this.invalidValue = invalidValue;
    }
}

function isKey(fieldSchema, indexName) {
    if (fieldSchema.type === 'Binary' ||
        fieldSchema.type === 'Custom' ||
        fieldSchema.type === 'Date' ||
        fieldSchema.type === 'Number' ||
        fieldSchema.type === 'String') {
        return indexName !== undefined
            ? Boolean(fieldSchema.indexKeyConfigurations &&
                fieldSchema.indexKeyConfigurations[indexName]) : Boolean(fieldSchema.keyType);
    }
    return false;
}

function keysFromSchema(schema) {
    const attributes = {};
    const tableKeys = {};
    const indexKeys = {};
    for (const propertyName of Object.keys(schema)) {
        const fieldSchema = schema[propertyName];
        if (fieldSchema.type === 'Binary' ||
            fieldSchema.type === 'Custom' ||
            fieldSchema.type === 'Date' ||
            fieldSchema.type === 'Number' ||
            fieldSchema.type === 'String') {
            const { attributeName = propertyName } = fieldSchema;
            if (fieldSchema.keyType) {
                attributes[attributeName] = attributeType(fieldSchema);
                tableKeys[attributeName] = fieldSchema.keyType;
            }
            if (fieldSchema.indexKeyConfigurations &&
                Object.keys(fieldSchema.indexKeyConfigurations).length > 0) {
                attributes[attributeName] = attributeType(fieldSchema);
                for (const indexName of Object.keys(fieldSchema.indexKeyConfigurations)) {
                    if (!(indexName in indexKeys)) {
                        indexKeys[indexName] = {};
                    }
                    indexKeys[indexName][attributeName]
                        = fieldSchema.indexKeyConfigurations[indexName];
                }
            }
        }
    }
    return { attributes, tableKeys, indexKeys };
}
function attributeType(fieldSchema) {
    switch (fieldSchema.type) {
        case 'Binary':
            return 'B';
        case 'Custom':
            if (!fieldSchema.attributeType) {
                throw new Error('Invalid schema: no attribute type defined for custom field');
            }
            return fieldSchema.attributeType;
        case 'Date':
        case 'Number':
            return 'N';
        case 'String':
            return 'S';
    }
}

function toSchemaName(path, schema) {
    if (typeof path === 'string') {
        path = new AttributePath(path);
    }
    const elements = path.elements.map(el => (Object.assign({}, el)));
    let cursor = {
        type: 'Document',
        members: schema
    };
    for (const element of elements) {
        if (element.type === 'AttributeName' &&
            cursor &&
            cursor.type === 'Document') {
            const { name } = element;
            element.name = getSchemaName(name, cursor.members);
            cursor = cursor.members[name];
        }
        else if (element.type === 'ListIndex' &&
            cursor &&
            cursor.type === 'List') {
            cursor = cursor.memberType;
        }
        else {
            break;
        }
    }
    return new AttributePath(elements);
}
function getSchemaName(propertyName, schema) {
    const fieldSchema = schema[propertyName];
    if (fieldSchema) {
        const { attributeName = propertyName } = fieldSchema;
        return attributeName;
    }
    return propertyName;
}

/**
 * Serialize a condition expression, substituting any property names for the
 * corresponding attribute names in the provided schema.
 *
 * @param expression The expression object to marshall.
 * @param schema The schema of the table to which the expression pertains.
 * @param attributes An optional ExpressionAttributes object to synchronize
 *                      substitutions across multiple expressions.
 */
function marshallConditionExpression(expression, schema, attributes = new ExpressionAttributes) {
    const serialized = serializeConditionExpression(normalizeConditionExpression(expression, schema), attributes);
    return {
        expression: serialized,
        ExpressionAttributeNames: attributes.names,
        ExpressionAttributeValues: attributes.values,
    };
}
/**
 * Serialize a function expression, substituting any property names for the
 * corresponding attribute names in the provided schema.
 *
 * @param expression The expression object to marshall.
 * @param schema The schema of the table to which the expression pertains.
 * @param attributes An optional ExpressionAttributes object to synchronize
 *                      substitutions across multiple expressions.
 */
function marshallFunctionExpression(expression, schema, attributes = new ExpressionAttributes) {
    const serialized = normalizeFunctionExpression(expression, schema)
        .serialize(attributes);
    return {
        expression: serialized,
        ExpressionAttributeNames: attributes.names,
        ExpressionAttributeValues: attributes.values,
    };
}
/**
 * Serialize a mathematical expression, substituting any property names for the
 * corresponding attribute names in the provided schema.
 *
 * @param expression The expression object to marshall.
 * @param schema The schema of the table to which the expression pertains.
 * @param attributes An optional ExpressionAttributes object to synchronize
 *                      substitutions across multiple expressions.
 */
function marshallMathematicalExpression(expression, schema, attributes = new ExpressionAttributes) {
    const serialized = normalizeMathematicalExpression(expression, schema)
        .serialize(attributes);
    return {
        expression: serialized,
        ExpressionAttributeNames: attributes.names,
        ExpressionAttributeValues: attributes.values,
    };
}
/**
 * Serialize a projection expression, substituting any property names for the
 * corresponding attribute names in the provided schema.
 *
 * @param expression The expression object to marshall.
 * @param schema The schema of the table to which the expression pertains.
 * @param attributes An optional ExpressionAttributes object to synchronize
 *                      substitutions across multiple expressions.
 */
function marshallProjectionExpression(expression, schema, attributes = new ExpressionAttributes) {
    const serialized = serializeProjectionExpression(expression.map(el => toSchemaName(el, schema)), attributes);
    return {
        expression: serialized,
        ExpressionAttributeNames: attributes.names,
        ExpressionAttributeValues: attributes.values,
    };
}
/**
 * Serialize an update expression, substituting any property names for the
 * corresponding attribute names in the provided schema.
 *
 * @param expression The expression object to marshall.
 * @param schema The schema of the table to which the expression pertains.
 * @param attributes An optional ExpressionAttributes object to synchronize
 *                      substitutions across multiple expressions.
 */
function marshallUpdateExpression(expression, schema, attributes = new ExpressionAttributes) {
    const serialized = normalizeUpdateExpression(expression, schema)
        .serialize(attributes);
    return {
        expression: serialized,
        ExpressionAttributeNames: attributes.names,
        ExpressionAttributeValues: attributes.values,
    };
}
function normalizeConditionExpression(expression, schema) {
    if (FunctionExpression.isFunctionExpression(expression)) {
        return normalizeFunctionExpression(expression, schema);
    }
    switch (expression.type) {
        case 'Equals':
        case 'NotEquals':
        case 'LessThan':
        case 'LessThanOrEqualTo':
        case 'GreaterThan':
        case 'GreaterThanOrEqualTo':
            return Object.assign({}, expression, { subject: toSchemaName(expression.subject, schema), object: normalizeIfPath(expression.object, schema) });
        case 'Between':
            return Object.assign({}, expression, { subject: toSchemaName(expression.subject, schema), lowerBound: normalizeIfPath(expression.lowerBound, schema), upperBound: normalizeIfPath(expression.upperBound, schema) });
        case 'Membership':
            return Object.assign({}, expression, { subject: toSchemaName(expression.subject, schema), values: expression.values.map(arg => normalizeIfPath(arg, schema)) });
        case 'Not':
            return Object.assign({}, expression, { condition: normalizeConditionExpression(expression.condition, schema) });
        case 'And':
        case 'Or':
            return Object.assign({}, expression, { conditions: expression.conditions.map(condition => normalizeConditionExpression(condition, schema)) });
    }
}
function normalizeFunctionExpression(expression, schema) {
    return new FunctionExpression(expression.name, ...expression.args.map(arg => normalizeIfPath(arg, schema)));
}
function normalizeMathematicalExpression(expression, schema) {
    return new MathematicalExpression(AttributePath.isAttributePath(expression.lhs) || typeof expression.lhs === 'string'
        ? toSchemaName(expression.lhs, schema)
        : expression.lhs, expression.operator, AttributePath.isAttributePath(expression.rhs) || typeof expression.rhs === 'string'
        ? toSchemaName(expression.rhs, schema)
        : expression.rhs);
}
const mapsToTransform = [
    ['toAdd', 'add'],
    ['toDelete', 'delete'],
    ['toSet', 'set'],
];
function normalizeUpdateExpression(expression, schema) {
    const normalized = new UpdateExpression;
    for (const [dataSet, exprMethod] of mapsToTransform) {
        for (const [path, value] of expression[dataSet]) {
            normalized[exprMethod](toSchemaName(path, schema), value);
        }
    }
    expression.toRemove.forEach(el => normalized.remove(toSchemaName(el, schema)));
    return normalized;
}
function normalizeIfPath(path, schema) {
    if (AttributePath.isAttributePath(path)) {
        return toSchemaName(path, schema);
    }
    return path;
}

const bytes = require('utf8-bytes');
/**
 * Converts a JavaScript object into a DynamoDB Item.
 *
 * @param schema Metadata explaining how the provided input is to be marshalled
 * @param input JavaScript object to convert
 */
function marshallItem(schema, input) {
    const marshalled = {};
    for (const key of Object.keys(schema)) {
        const value = input[key];
        const { attributeName = key } = schema[key];
        const marshalledValue = marshallValue(schema[key], value);
        if (marshalledValue) {
            marshalled[attributeName] = marshalledValue;
        }
    }
    return marshalled;
}
/**
 * Converts a value into a DynamoDB AttributeValue.
 *
 * @param schemaType    Metadata outlining how the value is to be understood and
 *                      converted
 * @param input         Value to convert
 */
function marshallValue(schemaType, input) {
    if (input === undefined) {
        const { defaultProvider } = schemaType;
        if (typeof defaultProvider === 'function') {
            input = defaultProvider();
        }
        else {
            return undefined;
        }
    }
    if (schemaType.type === 'Any') {
        const { onEmpty = 'nullify', onInvalid = 'omit', unwrapNumbers = false, } = schemaType;
        const marshaller = new Marshaller({ onEmpty, onInvalid, unwrapNumbers });
        return marshaller.marshallValue(input);
    }
    if (schemaType.type === 'Binary') {
        if (!input || input.length === 0 || input.byteLength === 0) {
            return { NULL: true };
        }
        return { B: marshallBinary(input) };
    }
    if (schemaType.type === 'Boolean') {
        return { BOOL: Boolean(input) };
    }
    if (schemaType.type === 'Custom') {
        return schemaType.marshall(input);
    }
    if (schemaType.type === 'Collection') {
        const { onEmpty = 'nullify', onInvalid = 'omit', unwrapNumbers = false, } = schemaType;
        const marshaller = new Marshaller({ onEmpty, onInvalid, unwrapNumbers });
        const collected = [];
        for (const element of input) {
            const marshalled = marshaller.marshallValue(element);
            if (marshalled) {
                collected.push(marshalled);
            }
        }
        return { L: collected };
    }
    if (schemaType.type === 'Date') {
        let date;
        if (typeof input === 'string') {
            date = new Date(input);
        }
        else if (typeof input === 'number') {
            date = new Date(input * 1000);
        }
        else if (isDate(input)) {
            date = input;
        }
        else {
            throw new InvalidValueError(input, 'Unable to convert value to date');
        }
        return { N: marshallNumber(Math.floor(date.valueOf() / 1000)) };
    }
    if (schemaType.type === 'Document') {
        return { M: marshallItem(schemaType.members, input) };
    }
    if (schemaType.type === 'Hash') {
        const { onEmpty = 'nullify', onInvalid = 'omit', unwrapNumbers = false, } = schemaType;
        const marshaller = new Marshaller({ onEmpty, onInvalid, unwrapNumbers });
        return { M: marshaller.marshallItem(input) };
    }
    if (schemaType.type === 'List') {
        const elements = [];
        for (const member of input) {
            const marshalled = marshallValue(schemaType.memberType, member);
            if (marshalled) {
                elements.push(marshalled);
            }
        }
        return { L: elements };
    }
    if (schemaType.type === 'Map') {
        const marshalled = {};
        if (typeof input[Symbol.iterator] === 'function') {
            for (let [key, value] of input) {
                const marshalledValue = marshallValue(schemaType.memberType, value);
                if (marshalledValue) {
                    marshalled[key] = marshalledValue;
                }
            }
        }
        else if (typeof input === 'object') {
            for (const key of Object.keys(input)) {
                const marshalledValue = marshallValue(schemaType.memberType, input[key]);
                if (marshalledValue) {
                    marshalled[key] = marshalledValue;
                }
            }
        }
        else {
            throw new InvalidValueError(input, 'Unable to convert value to map');
        }
        return { M: marshalled };
    }
    if (schemaType.type === 'Null') {
        return { NULL: true };
    }
    if (schemaType.type === 'Number') {
        return { N: marshallNumber(input) };
    }
    if (schemaType.type === 'Set') {
        if (schemaType.memberType === 'Binary') {
            if (!(input instanceof BinarySet)) {
                const set = new BinarySet();
                for (const item of input) {
                    set.add(marshallBinary(item));
                }
                input = set;
            }
            return marshallSet(input, marshallBinary, (bin) => bin.byteLength === 0, 'BS');
        }
        if (schemaType.memberType === 'Number') {
            if (!(input instanceof Set)) {
                input = new NumberValueSet(input);
            }
            return marshallSet(input, marshallNumber, () => false, 'NS');
        }
        if (schemaType.memberType === 'String') {
            if (!(input instanceof Set)) {
                const original = input;
                input = new Set();
                for (const el of original) {
                    input.add(el);
                }
            }
            return marshallSet(input, marshallString, (string) => string.length === 0, 'SS');
        }
        throw new InvalidSchemaError(schemaType, `Unrecognized set member type: ${schemaType.memberType}`);
    }
    if (schemaType.type === 'String') {
        const string = marshallString(input);
        if (string.length === 0) {
            return { NULL: true };
        }
        return { S: string };
    }
    if (schemaType.type === 'Tuple') {
        return {
            L: schemaType.members
                .map((type, index) => marshallValue(type, input[index]))
                .filter((val) => val !== undefined)
        };
    }
    throw new InvalidSchemaError(schemaType, 'Unrecognized schema node');
}
function marshallBinary(input) {
    if (ArrayBuffer.isView(input)) {
        return new Uint8Array(input.buffer, input.byteOffset, input.byteLength);
    }
    if (isArrayBuffer(input)) {
        return new Uint8Array(input);
    }
    return Uint8Array.from(bytes(input));
}
function marshallNumber(input) {
    return input.toString(10);
}
function marshallString(input) {
    return input.toString();
}
function marshallSet(value, marshaller, isEmpty, setTag) {
    const collected = [];
    for (const member of value) {
        const marshalled = marshaller(member);
        if (isEmpty(marshalled)) {
            // DynamoDB sets cannot contain empty values
            continue;
        }
        collected.push(marshalled);
    }
    if (collected.length === 0) {
        return { NULL: true };
    }
    return { [setTag]: collected };
}
function isArrayBuffer(arg) {
    return typeof ArrayBuffer === 'function'
        && (arg instanceof ArrayBuffer ||
            Object.prototype.toString.call(arg) === '[object ArrayBuffer]');
}
function isDate(arg) {
    return arg instanceof Date
        || Object.prototype.toString.call(arg) === '[object Date]';
}

function marshallKey(schema, input, indexName) {
    const marshalled = {};
    for (const propertyKey of Object.keys(schema)) {
        const fieldSchema = schema[propertyKey];
        if (isKey(fieldSchema, indexName)) {
            const { attributeName = propertyKey } = fieldSchema;
            const value = marshallValue(fieldSchema, input[propertyKey]);
            if (value) {
                marshalled[attributeName] = value;
            }
        }
    }
    return marshalled;
}

/**
 * The enumeration of types supported by this marshaller package.
 */
const TypeTags = {
    Any: 'Any',
    Binary: 'Binary',
    Boolean: 'Boolean',
    Collection: 'Collection',
    Custom: 'Custom',
    Date: 'Date',
    Document: 'Document',
    Hash: 'Hash',
    List: 'List',
    Map: 'Map',
    Null: 'Null',
    Number: 'Number',
    Set: 'Set',
    String: 'String',
    Tuple: 'Tuple',
};
function isBaseType(arg) {
    return Boolean(arg) && typeof arg === 'object'
        && typeof arg.type === 'string'
        && arg.type in TypeTags
        && ['string', 'undefined'].indexOf(typeof arg.attributeName) > -1;
}
/**
 * The types of keys a given attribute can represent.
 */
const KeyTypes = {
    HASH: 'HASH',
    RANGE: 'RANGE',
};
function isKeyableType(arg) {
    const { keyType, indexKeyConfigurations } = arg;
    if (!(keyType === undefined || keyType in KeyTypes)) {
        return false;
    }
    const idxKeysType = typeof indexKeyConfigurations;
    if (indexKeyConfigurations && idxKeysType === 'object') {
        for (const indexName of Object.keys(indexKeyConfigurations)) {
            if (!(indexKeyConfigurations[indexName] in KeyTypes)) {
                return false;
            }
        }
        return true;
    }
    return idxKeysType === 'undefined';
}
function isSchemaType(arg, alreadyVisited = new Set()) {
    if (isBaseType(arg)) {
        if (alreadyVisited.has(arg)) {
            return true;
        }
        alreadyVisited.add(arg);
        switch (arg.type) {
            case 'Binary':
            case 'Date':
            case 'String':
                return isKeyableType(arg);
            case 'Custom':
                return isKeyableType(arg)
                    && typeof arg.marshall === 'function'
                    && typeof arg.unmarshall === 'function'
                    && [
                        void 0,
                        'S',
                        'N',
                        'B',
                    ].indexOf(arg.attributeType) > -1;
            case 'Document':
                return isDocumentType(arg, alreadyVisited);
            case 'List':
            case 'Map':
                return isSchemaType(arg.memberType, alreadyVisited);
            case 'Number':
                return isKeyableType(arg) && ['boolean', 'undefined']
                    .indexOf(typeof arg.versionAttribute) > -1;
            case 'Tuple':
                return isTupleType(arg, alreadyVisited);
            default:
                return true;
        }
    }
    return false;
}
function isDocumentType(arg, alreadyVisited) {
    const { valueConstructor, members } = arg;
    if (!members || typeof members !== 'object') {
        return false;
    }
    for (let key of Object.keys(members)) {
        if (!isSchemaType(members[key], alreadyVisited)) {
            return false;
        }
    }
    return ['function', 'undefined',].indexOf(typeof valueConstructor) > -1;
}
function isTupleType(arg, alreadyVisited) {
    const { members } = arg;
    if (!Array.isArray(members)) {
        return false;
    }
    for (let member of members) {
        if (!isSchemaType(member, alreadyVisited)) {
            return false;
        }
    }
    return true;
}

/**
 * Evaluates whether the provided argument is a Schema object
 */
function isSchema(arg) {
    if (!Boolean(arg) || typeof arg !== 'object') {
        return false;
    }
    for (let key of Object.keys(arg)) {
        if (!isSchemaType(arg[key])) {
            return false;
        }
    }
    return true;
}

/**
 * Unmarshall a DynamoDB item into a JavaScript value.
 *
 * @param schema            Metadata outlining the types to be expected
 *                          throughout the input
 * @param input             The value to unmarshall
 * @param valueConstructor  A zero-argument constructor used to create the
 *                          object onto which the input should be unmarshalled
 */
function unmarshallItem(schema, input, valueConstructor) {
    const unmarshalled = valueConstructor
        ? new valueConstructor()
        : Object.create(null);
    for (const key of Object.keys(schema)) {
        const { attributeName = key } = schema[key];
        if (attributeName in input) {
            unmarshalled[key] = unmarshallValue(schema[key], input[attributeName]);
        }
    }
    return unmarshalled;
}
function unmarshallValue(schemaType, input) {
    switch (schemaType.type) {
        case 'Any':
        case 'Collection':
        case 'Hash':
            const autoMarshaller = new Marshaller();
            return autoMarshaller.unmarshallValue(input);
        case 'Binary':
            if (input.NULL) {
                return new Uint8Array(0);
            }
            return input.B;
        case 'Boolean':
            return input.BOOL;
        case 'Custom':
            return schemaType.unmarshall(input);
        case 'Date':
            return input.N ? new Date(Number(input.N) * 1000) : undefined;
        case 'Document':
            return input.M
                ? unmarshallItem(schemaType.members, input.M, schemaType.valueConstructor) : undefined;
        case 'List':
            return input.L ? unmarshallList(schemaType, input.L) : undefined;
        case 'Map':
            return input.M ? unmarshallMap(schemaType, input.M) : undefined;
        case 'Null':
            return input.NULL ? null : undefined;
        case 'Number':
            return typeof input.N === 'string' ? Number(input.N) : undefined;
        case 'Set':
            switch (schemaType.memberType) {
                case 'Binary':
                    if (input.NULL) {
                        return new BinarySet();
                    }
                    return typeof input.BS !== 'undefined'
                        ? new BinarySet(input.BS)
                        : undefined;
                case 'Number':
                    if (input.NULL) {
                        return new Set();
                    }
                    return input.NS ? unmarshallNumberSet(input.NS) : undefined;
                case 'String':
                    if (input.NULL) {
                        return new Set();
                    }
                    return input.SS ? unmarshallStringSet(input.SS) : undefined;
                default:
                    throw new InvalidSchemaError(schemaType, `Unrecognized set member type: ${schemaType.memberType}`);
            }
        case 'String':
            return input.NULL ? '' : input.S;
        case 'Tuple':
            return input.L ? unmarshallTuple(schemaType, input.L) : undefined;
    }
    throw new InvalidSchemaError(schemaType, 'Unrecognized schema node');
}
function unmarshallList(schemaType, input) {
    const list = [];
    for (const element of input) {
        list.push(unmarshallValue(schemaType.memberType, element));
    }
    return list;
}
function unmarshallMap(schemaType, input) {
    const map = new Map();
    for (const key of Object.keys(input)) {
        map.set(key, unmarshallValue(schemaType.memberType, input[key]));
    }
    return map;
}
function unmarshallNumberSet(input) {
    const set = new Set();
    for (const number of input) {
        set.add(Number(number));
    }
    return set;
}
function unmarshallStringSet(input) {
    const set = new Set();
    for (const string of input) {
        set.add(string);
    }
    return set;
}
function unmarshallTuple(schemaType, input) {
    const { members } = schemaType;
    const tuple = [];
    for (let i = 0; i < members.length; i++) {
        tuple.push(unmarshallValue(members[i], input[i]));
    }
    return tuple;
}

export { InvalidSchemaError, InvalidValueError, isKey, keysFromSchema, marshallConditionExpression, marshallFunctionExpression, marshallMathematicalExpression, marshallProjectionExpression, marshallUpdateExpression, marshallItem, marshallValue, marshallKey, isSchema, TypeTags, KeyTypes, isSchemaType, toSchemaName, getSchemaName, unmarshallItem };
